<?php

namespace Goldenscarab\Modulus\Debusca\App\Http\Drivers;

use PDO;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Cache;
use Goldenscarab\Modulus\Debusca\App\Http\Drivers\Objects\Meta;
use Goldenscarab\Modulus\Debusca\App\Http\Drivers\Objects\MetaCollect;
use Goldenscarab\Modulus\Debusca\App\Http\Drivers\Objects\QueryResult;


class Mysql extends Driver
{
    private $queries = [
        'tables'  => "SHOW TABLES;",
        'schema'  => "select * from INFORMATION_SCHEMA.COLUMNS WHERE `TABLE_SCHEMA` = '{BDD}'",
        'columns' => "SELECT ORDINAL_POSITION AS POSITION, COLUMN_NAME AS 'NAME', COLUMN_TYPE AS 'TYPE', CASE COLUMN_KEY WHEN 'PRI' THEN 'PRIMARY' WHEN 'UNI' THEN 'UNIQUE' ELSE NULL END AS 'KEY', COLUMN_DEFAULT AS 'DEFAULT', IS_NULLABLE, EXTRA  from INFORMATION_SCHEMA.COLUMNS WHERE `TABLE_SCHEMA` = '{BDD}' AND `TABLE_NAME` = '{TABLE}'",
        'select'  => "SELECT * FROM `{TABLE}` WHERE 1 LIMIT 25 OFFSET 0;",
        'insert'  => "INSERT INTO `{TABLE}` ({COLUMNS_NAMES}) VALUES ({COLUMNS_VALUES});",
        'update'  => "UPDATE `{TABLE}` AS {ALIAS} SET {NAMES_VALUES} WHERE {ALIAS}.`{PK_NAME}` = {ID};",
        'delete'  => "DELETE FROM `{TABLE}` WHERE `{PK_NAME}` = {ID};",
        'create'  => "CREATE TABLE `my_table` (`id` INT(10) NOT NULL, `name` VARCHAR(255) NOT NULL, `status` TINYINT(1) NOT NULL DEFAULT '0', `date` DATE NULL, PRIMARY KEY (`id`)) ENGINE = InnoDB;",
        'alter'   => "ALTER TABLE `{TABLE}` ADD `[COL_NAME]` varchar(255) NOT NULL AFTER [OTHER_COL];\nALTER TABLE `{TABLE}` CHANGE `[COL_OLD_NAME]` `[COL_NEW_NAME]` VARCHAR(255) CHARACTER SET utf8 COLLATE utf8_unicode_ci NOT NULL;\nALTER TABLE `{TABLE}` MODIFY COLUMN [COL_NAME] varchar(255) NOT NULL AFTER [OTHER_COL_NAME];"
    ];

    private $tables_metas = null;

    public function __construct($db)
    {
        parent::__construct($db);

        foreach ($this->queries as $key => $query) {
            $this->queries[$key] = str_replace('{BDD}', $this->db->getConfig('database'), $query);
        }
    }

    public function getTables(): Collection
    {
        $cache_key = 'mysql.' . $this->db->getConfig('name') . '.tables';

        $tables = Cache::remember($cache_key, 5 * 60, function () {

            $tables = collect();

            $statement = $this->db->getPdo()->prepare($this->queries['tables']);
            $statement->execute();
            $result = $statement->fetchAll(PDO::FETCH_NUM);

            $tables = collect($result)->map(function($item) {
                return data_get($item, 0);
            });

            $collect = collect();
            foreach ($tables as $table) {
                $metas = $this->getMetas($table);
                $qr = new QueryResult($metas, [$table], $this->queries);
                $collect->push($qr);
            }

            return $collect;

        });

        return $tables;
    }

    public function getMetas($table): Collection
    {
        $cache_key = 'mysql.' . $this->db->getConfig('name') . '.metas';

        $tables_metas = &$this->tables_metas;

        $metas = Cache::remember($cache_key, 5 * 60, function () use ($table, $tables_metas) {

            if (is_null($tables_metas)) {
                //$query = str_replace('{BDD}', $this->db->getConfig('database'), $this->queries['schema']);
                $tables_metas = collect($this->db->select($this->queries['schema']));
            }

            $columns = $tables_metas->where('TABLE_NAME', $table);

            $metas = new MetaCollect();

            foreach ($columns as $column) {

                $key = null;
                if ($column->COLUMN_KEY == 'PRI') {
                    $key = 'primary_key';
                } else if ($column->COLUMN_KEY == 'UNI') {
                    $key = 'unique';
                }
                $meta = new Meta;
                $meta->setTable($table);
                $meta->setField($column->COLUMN_NAME);
                $meta->setType($column->DATA_TYPE);
                $meta->setNull($column->IS_NULLABLE == 'NO' ? false : true);
                $meta->setKey($key);
                $meta->setDefault($column->COLUMN_DEFAULT);
                $meta->setExtra($column->EXTRA);

                $metas->push($meta);
            }

            return $metas;
        });

        return $metas;
    }

    public function getSuggestions(): Collection
    {
        $tables_metas = &$this->tables_metas;
        $cache_key = 'mysql.' . $this->db->getConfig('name') . '.suggestions';

        $suggestions = Cache::remember($cache_key, 5 * 60, function () use ($tables_metas) {

            $suggestions = collect();

            if (is_null($tables_metas)) {
                $tables_metas = $this->db->select($this->queries['schema']);
            }

            foreach ($tables_metas as $column) {
                $suggestions->push($column->TABLE_NAME);
                $suggestions->push($column->COLUMN_NAME);
                $suggestions->push($column->TABLE_NAME . '.' . $column->COLUMN_NAME);
            }

            return $suggestions;
        });

        return $suggestions;
    }

    public function query($query): QueryResult
    {
        $time_start = microtime(true);
        $splited = explode(' ', strtolower($query));
        $type    = array_shift($splited);
        $result  = null;

        $_query = DB::raw($query);
        $statement = $this->db->getPdo()->prepare($_query);
        $status = $statement->execute();

        switch ($type) {
            case 'select':
            case 'describe':
            case 'show':
                $result = $statement->fetchAll(PDO::FETCH_ASSOC);
                break;
            case 'update':
            case 'delete':
            case 'exec':
                $result = $statement->rowCount();
                break;
            default:
                $result = $status;
        }
        $time_end = microtime(true);

        $meta = $this->getStatementMeta($statement);

        $qr = new QueryResult($meta, $result, $this->queries);
        $qr->setResume([
            'type'       => $type,
            'rows'       => $statement->rowCount(),
            'elapsed_ms' => round(($time_end - $time_start) * 1000, 1)
        ]);

        return $qr;

    }

    private function getStatementMeta($statement)
    {
        $metas = new MetaCollect();

        for ($i = 0; $i < $statement->columnCount(); $i++) {
            $sm = $statement->getColumnMeta($i);

            $key = '';
            if (in_array('primary_key', $sm['flags'])) {
                $key = 'primary_key';
            } else if (in_array('unique_key', $sm['flags'])) {
                $key = 'unique';
            }

            $meta = new Meta;
            $meta->setTable(data_get($sm, 'table'));
            $meta->setField(data_get($sm, 'name'));
            $meta->setType(data_get($sm, 'native_type'));
            $meta->setNull(in_array('not_null', data_get($sm, 'flags')) ? false : true);
            $meta->setKey($key);

            $metas->push($meta);
        }

        return $metas;
    }

    public function getQueryString($query)
    {
        return data_get($this->queries, $query);
    }


}
